"""Create a JLCPCB PCBA set of files to support PCBA, this requires us to
produce a BOM file and a CPL (component placement file), which will be a
.pos file.
We do this by reading the associated schematic (mainly for part numbers)
and then cross-matching the pcb modules.
"""
import pcbnew
import re
import os


def read_rotation_db(filename):
    """Read the cf config file turns so we know which turns to run later.
    """
    db = []
    fh = open(filename, 'r')
    for line in fh:
        line = line.rstrip(' \r\n')  # All newlines and spaces should be removed
        line = re.sub(r'\s\s+', ' ', line)  # Remove all double spaces
        if line == "":
            continue
        m = re.match(r"^(?P<Start>[^\s]+)\s+(?P<End>\d+)$", line)
        if m:
            db.append((m.group(1), int(m.group(2))))
    return db


def possible_rotate(footprint):
    """Given the footprint name, work out what rotation is needed, we
    support matching against the long or short footprint names (if there
    is a colon in the regex).
    """
    fpshort = footprint.split(':')[-1]
    for rot in rotation_db:
        ex = rot[0]
        delta = rot[1]
        fp = fpshort
        if re.search(r':', ex):
            fp = footprint
        if re.search(ex, fp):
            return delta
    return 0


def create_pcba():
    """Main function to create the files.
    """

    global rotation_db

    board = pcbnew.GetBoard()
    board_file = board.GetFileName()
    project_path = os.path.dirname(board_file)
    projet_name = os.path.splitext(os.path.basename(board_file))[0]
    # projet_name = ""

    # Populate the rotation db (do it here so editing and retrying is easy).
    rotation_db = read_rotation_db(os.path.join(os.path.dirname(__file__), 'rotation.properties'))
    # rotation_db = read_rotation_db(os.path.join(os.path.dirname(__file__), 'rotations.cf'))

    # Open both layer files...
    top_file = os.path.join(project_path, projet_name) + 'JLCPCB_CPL_file.csv'
    top_fh = open(top_file, 'w')
    top_fh.write('"Designator","Mid X","Mid Y","Rotation","Layer"\r\n')

    bom_list = {}
    for f in board.GetFootprints():

        # Check if it is a SMD component, PHT components are node assembled
        # by JCLPCB on the default production line.
        if not (f.GetAttributes() & pcbnew.FP_SMD) \
                or (f.GetAttributes() &
                    (pcbnew.FP_BOARD_ONLY + pcbnew.FP_EXCLUDE_FROM_POS_FILES + pcbnew.FP_BOARD_ONLY)):
            continue

        # Add item to the BOM creating a `set()`.
        reference = f.GetReference()
        value = f.GetValue().strip()
        # Check the typed name for the LCSC stock code filed name.
        lcsc_field_name = 'lcsc'
        for field_name in f.GetProperties().keys():
            if field_name.lower() == 'lcsc' or field_name.lower() == 'lcsc#':
                lcsc_field_name = field_name
                break
        lcsc_code = f.GetProperty(lcsc_field_name).strip()
        # Add compatibility logic with KiCost (https://github.com/xesscorp/KiCost)
        # by using the first valid LCSC stock code in case of multiple part assigned.
        lcsc_codes = lcsc_code.split(';')
        for c in lcsc_codes:
            c = c.strip()
            if c:
                lcsc_code = c
                break

        if not (re.search(r'\d$', reference) and re.search(r'\d', value)) and lcsc_field_name:
            # Die Referenz muss eine numerische Endung haben, normalerweise Pcbnew
            # Verwenden Sie G*** oder diese Art der Terminierung für Fußabdrücke, die
            # stellen keinen Teil dar.
            # Auch der Wert muss eine Zahl haben, normalerweise der Designer
            # Teile "nicht bevölkern" / "n/a" (entfällt) ... zuweisen
            # auf Layout, das später zusammengebaut werden kann.
            # Um eine gewisse Erwartung zu haben, falls ein LCSC-Aktiencode vorhanden ist,
            # dieser Teil wird berücksichtigt.
            continue

        try:
            footprint = str(f.GetFPID().GetLibItemName())
        except RuntimeError:
            footprint = None
        if not lcsc_code:
            lcsc_code = str(None)  # It will be selected after at JLCPCB interface.
        if reference and value and footprint:
            # footprint = footprint.split(':')[-1] # Simplify the footprint name.
            key = value + '//' + footprint + '//' + lcsc_code
            if key not in bom_list:
                bom_list[key] = set()
            bom_list[key].add(reference)

        # Get the position of the component. Internally, Pcbnew uses
        # micrometer position unit and a tenth of degrees on rotation.
        x = f.GetX() / 1000000.0
        y = - f.GetY() / 1000000.0

        # -tudo x & Y new calculate
        # U1  MCP6022SOIC-8-1EP_3.9x4.9mm_P1.27mm_EP2.41x3.3mm 21.8060 21.5000 90.0000top
        # L = 7.4mm
        # B = 5.4mm
        # X = 21.806mm
        # Y = 21.5mm
        # Mid X = 7.4 / 2 + 21.5 = 25.2
        # Mid Y = 5.4 / 2 + 21.806 = 24.506

        rotation = f.GetOrientationDegrees()  # Alredy give the converted value.
        rotation = (rotation + possible_rotate(footprint)) % 360

        # Write the component to the correspondent position file.
        # Use `f.GetLayer()` instead `f.GetLayerName()` because v6
        # will allow layer rename.

        layer_name = ('B' if f.GetLayer() & pcbnew.SIDE_TOP else 'T')
        top_fh.write('"' + reference + '","' + str(x) + 'mm","' + str(y) + 'mm","' + str(
            rotation) + '","' + layer_name + '"\r\n')

    top_fh.close()

    # Write the BOM.
    bom_file = os.path.join(project_path, projet_name) + 'JLCPCB_BOM_File.csv'
    bom_fh = open(bom_file, 'w')
    bom_fh.write('"Comment","Designator","Footprint","LCSC"\r\n')

    for k, v in bom_list.items():
        value, footprint, lcsc = re.split("//", k)
        references = ','.join(v)
        bom_fh.write('"' + value + '","' + references + '","' +
                     footprint + '","' + lcsc + '"\r\n')

    bom_fh.close()
